from torch import nn
from .utils import load_state_dict_from_url

__all__ = ['MobileNetV2', 'mobilenet_v2']

model_urls = {
    'mobilenet_v2': 'https://download.pytorch.org/models/mobilenet_v2-b0353104.pth',
}


def _make_divisible(v, divisor, min_value=None):
    """
    This function is taken from the original tf repo.
    It ensures that all layers have a channel number that is divisible by 8
    It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    :param v:
    :param divisor:
    :param min_value:
    :return:
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


class ConvBNReLU(nn.Sequential):
    def __init__(self, in_planes, out_planes, kernel_size=3, stride=1, groups=1):
        padding = (kernel_size - 1) // 2
        super(ConvBNReLU, self).__init__(
            nn.Conv2d(in_planes, out_planes, kernel_size, stride, padding, groups=groups, bias=False),
            nn.BatchNorm2d(out_planes),
            nn.ReLU6(inplace=True)
        )


class InvertedResidual(nn.Module):
    def __init__(self, inp, oup, stride, expand_ratio):
        super(InvertedResidual, self).__init__()
        self.stride = stride
        assert stride in [1, 2]

        hidden_dim = int(round(inp * expand_ratio))
        self.use_res_connect = self.stride == 1 and inp == oup

        layers = []
        if expand_ratio != 1:
            # pw
            layers.append(ConvBNReLU(inp, hidden_dim, kernel_size=1))
        layers.extend([
            # dw
            ConvBNReLU(hidden_dim, hidden_dim, stride=stride, groups=hidden_dim),
            # pw-linear
            nn.Conv2d(hidden_dim, oup, 1, 1, 0, bias=False),
            nn.BatchNorm2d(oup),
        ])
        self.conv = nn.Sequential(*layers)

    def forward(self, x):
        if self.use_res_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)


class MobileNetV2(nn.Module):
    def __init__(self,
                 num_classes=1000,
                 width_mult=1.0,
                 inverted_residual_setting=None,
                 round_nearest=8,
                 block=None):
        """
        MobileNet V2 main class

        Args:
            num_classes (int): Number of classes
            width_mult (float): Width multiplier - adjusts number of channels in each layer by this amount
            inverted_residual_setting: Network structure
            round_nearest (int): Round the number of channels in each layer to be a multiple of this number
            Set to 1 to turn off rounding
            block: Module specifying inverted residual building block for mobilenet

        """
        super(MobileNetV2, self).__init__()

        if block is None:
            block = InvertedResidual
        input_channel = 32
        last_channel = 1280

        if inverted_residual_setting is None:
            inverted_residual_setting = [
                # t, c, n, s
                [1, 16, 1, 1],
                [6, 24, 2, 2],
                [6, 32, 3, 2],
                [6, 64, 4, 2],
                [6, 96, 3, 1],
                [6, 160, 3, 2],
                [6, 320, 1, 1],
            ]

        # only check the first element, assuming user knows t,c,n,s are required
        if len(inverted_residual_setting) == 0 or len(inverted_residual_setting[0]) != 4:
            raise ValueError("inverted_residual_setting should be non-empty "
                             "or a 4-element list, got {}".format(inverted_residual_setting))

        # building first layer
        input_channel = _make_divisible(input_channel * width_mult, round_nearest)
        self.last_channel = _make_divisible(last_channel * max(1.0, width_mult), round_nearest)
        features = [ConvBNReLU(3, input_channel, stride=2)]
        # building inverted residual blocks
        for t, c, n, s in inverted_residual_setting:
            output_channel = _make_divisible(c * width_mult, round_nearest)
            for i in range(n):
                stride = s if i == 0 else 1
                features.append(block(input_channel, output_channel, stride, expand_ratio=t))
                input_channel = output_channel
        # building last several layers
        features.append(ConvBNReLU(input_channel, self.last_channel, kernel_size=1))
        # make it nn.Sequential
        self.features = nn.Sequential(*features)

        # building classifier
        self.classifier = nn.Sequential(
            nn.Dropout(0.2),
            nn.Linear(self.last_channel, num_classes),
        )

        # weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out')
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.zeros_(m.bias)

    def _forward(self, x):
        x = self.features(x)
        x = x.mean([2, 3])
        x = self.classifier(x)
        return x

    # Allow for accessing forward method in a inherited class
    forward = _forward


def mobilenet_v2(pretrained=False, progress=True, **kwargs):
    """
    Constructs a MobileNetV2 architecture from
    `"MobileNetV2: Inverted Residuals and Linear Bottlenecks" <https://arxiv.org/abs/1801.04381>`_.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    model = MobileNetV2(**kwargs)
    if pretrained:
        state_dict = load_state_dict_from_url(model_urls['mobilenet_v2'],
                                              progress=progress)
        model.load_state_dict(state_dict)
    return model



import gc
import inspect

import numpy as np
import torch

from yqn_pytorch_framework.device import tensor_load_device


def read_frame(frame):
    func_name = frame.f_code.co_name
    filename = frame.f_globals["__file__"]
    if (filename.endswith(".pyc") or
            filename.endswith(".pyo")):
        filename = filename[:-1]
    module_name = frame.f_globals["__name__"]
    curr_line = frame.f_lineno
    return func_name, module_name, curr_line


def get_tensors():
    tensor_list = []
    for obj in gc.get_objects():
        try:
            if torch.is_tensor(obj) or (hasattr(obj, 'data') and torch.is_tensor(obj.data)):
                tensor = obj
            else:
                continue
            if tensor.is_cuda:
                tensor_list.append(tensor)
        except Exception as e:
            pass
            # print('A trivial exception occurred: {}'.format(e))
    return tensor_list


def load_current_tensor():
    tensor_list = get_tensors()
    tensor_size_list = [tensor.size() for tensor in tensor_list]
    tensor_sizes = {(type(x),
                     tuple(x.size()),
                     tensor_size_list.count(x.size()),
                     np.prod(np.array(x.size())) * 4 / 1000 ** 2) for x in tensor_list}
    return tensor_sizes


def _track_step(frame, device):
    import pynvml
    pynvml.nvmlInit()
    handle = pynvml.nvmlDeviceGetHandleByIndex(device.index)
    mem_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
    # func_name, module_name, curr_line = read_frame(frame)
    # where_str = module_name + ' ' + func_name + ':' + ' line ' + str(curr_line)
    mem_info_mb = mem_info.used / 1000 ** 2
    tensor_list_with_sizes = load_current_tensor()
    tensor_mem_size = 0
    for tensor in tensor_list_with_sizes:
        tensor_mem_size += tensor[3]
    pynvml.nvmlShutdown()
    return "", mem_info_mb, tensor_list_with_sizes, tensor_mem_size


def model_size_estimate(device, get_model):
    _gpu_real_trace_size(device, get_model)
    from yqn_pytorch_framework.model.base_model import BaseModel
    model = tensor_load_device(get_model(), device).eval()
    if isinstance(model, BaseModel):
        _model_size(model, model.get_default_input(), device)


def _gpu_real_trace_size(device, get_model):
    # print(torch.cuda.memory_summary(device))
    frame = inspect.currentframe()
    where_str, preheat_mem_info_mb, preheat_tensor_size, _ = _track_step(frame, device)
    dummy_tensor_1 = tensor_load_device(torch.randn(1, 10, 100, 100).float(), device)
    where_str, base_mem_info_mb, base_tensor_size, _ = _track_step(frame, device)
    # print(f"GPU Memory Track | {datetime.datetime.now():%d-%b-%y-%H:%M:%S} |"
    #       f" Total Used Memory:{base_mem_info_mb:<7.1f}Mb\n\n")
    model = tensor_load_device(get_model(), device).eval()
    where_str, after_model_mem_info_mb, after_model_tensor_sizes, _ = _track_step(frame, device)
    # for t, s, n, m in after_model_tensor_sizes - base_tensor_size:
    #     print(f'add     | {str(n)} * Size:{str(s):<20} | Memory: {str(m * n)[:6]} M | {str(t):<20}\n')
    # for t, s, n, m in base_tensor_size - after_model_tensor_sizes:
    #     print(f'release | {str(n)} * Size:{str(s):<20} | Memory: {str(m * n)[:6]} M | {str(t):<20}\n')
    # print(f"GPU Memory Track | {datetime.datetime.now():%d-%b-%y-%H:%M:%S} |"
    #       f" Total Used Memory:{after_model_mem_info_mb:<7.1f}Mb\n\n")
    print(f"{model._get_name()} GPU Frame {(after_model_mem_info_mb - base_mem_info_mb):<7.1f}Mb")
    print(torch.cuda.memory_summary(device))
    del model, dummy_tensor_1
    torch.cuda.empty_cache()


def _model_size(model, input_features, device, type_size=4):
    """
    数据占用
    参数占用
    :param model:
    :param input_features:
    :param type_size:
    :return:
    """
    frame = inspect.currentframe()
    model_name = model._get_name()
    with torch.no_grad():
        para = sum([np.prod(list(p.size())) for p in model.parameters()])
        print('{} Params Used: {:<7.1f}M'.format(model_name, para * type_size / 1000 / 1000))
        # if input_features is tuple:
        #     input_ = input_features.clone()
        #     input_.requires_grad_(requires_grad=False)
        # else:
        #     input_ = input_features.clone()
        #     input_.requires_grad_(requires_grad=False)

        where_str, preheat_mem_info_mb, preheat_tensor_list, preheat_tensor_size = _track_step(frame, device)
        where_str, base_mem_info_mb, base_tensor_list, base_tensor_size = _track_step(frame, device)
        result = model(input_features)
        where_str, after_model_mem_info_mb, after_model_tensor_list, after_model_tensor_size = _track_step(frame,
                                                                                                           device)

        print(f"{model_name} Variables Used {(after_model_mem_info_mb - base_mem_info_mb):<7.1f}Mb")
    print(torch.cuda.memory_summary(device))
    del model
    torch.cuda.empty_cache()

    # mods = list(model.compute_modules())
    # out_sizes = []
    #
    # for i in range(1, len(mods)):
    #     m = mods[i]
    #     if isinstance(m, nn.ReLU):
    #         if m.inplace:
    #             continue
    #     out = m(input_)
    #     out_sizes.append(np.array(out.size()))
    #     input_ = out
    #
    # total_nums = 0
    # for i in range(len(out_sizes)):
    #     s = out_sizes[i]
    #     nums = np.prod(np.array(s))
    #     total_nums += nums
    # print('{} : intermediates variables: {:3f} M (infer)'
    #       .format(model_name, total_nums * type_size / 1000 / 1000))
    # print('{} : intermediates variables: {:3f} M (train)'
    #       .format(model_name, total_nums * type_size * 2 / 1000 / 1000))



class NoneException(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class InputSizeException(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class InferModelException(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ConfigMissException(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class LabelDirException(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)



from configparser import ConfigParser


class CommonConf(ConfigParser):
    def __init__(self, defaults=None):
        ConfigParser.__init__(self, defaults=defaults)
        self.add_sec = "Additional"

    def optionxform(self, option_str):
        return option_str



import time

import torch

from yqn_pytorch_framework.device import tensor_load_device


def model_time_estimate(device, get_model, loop=100):
    model = tensor_load_device(get_model(), device).eval()
    torch.cuda.synchronize(device)
    start = time.time()
    with torch.no_grad():
        for i in range(loop):
            result = model(model.get_default_input())
    torch.cuda.synchronize(device)
    end = time.time()
    print(f"{model._get_name()} GPU PerLoop Cost {(end - start) * 1000 / loop:<5.2f}ms")



import os
import re

import oss2
import sys

import yaml

HASH_REGEX = re.compile(r'-([a-f0-9]*)\.')
HOST_URL = "https://ai-models-yqn.oss-cn-hangzhou.aliyuncs.com"


def oss_calculate_file_crc64(file_name, block_size=64 * 1024, init_crc=0):
    """计算文件的MD5
    :param file_name: 文件名
    :param block_size: 计算MD5的数据块大小，默认64KB
    :return 文件内容的MD5值
    """
    with open(file_name, 'rb') as f:
        crc64 = oss2.utils.Crc64(init_crc)
        while True:
            data = f.read(block_size)
            if not data:
                break
            crc64.update(data)

    return crc64.crc


class OSSUpload:
    def __init__(self, app_id, version, bucket):
        self.app_id = app_id
        self.version = version
        self.bucket = bucket

    def upload_dir(self, dir_path):
        dist_start_path = os.path.join(self.app_id, self.version).__str__()
        self.rm_exist_version(dist_start_path)
        outputs = []
        self.add_file_to_oss(dir_path, dist_start_path, outputs)
        with open(r'./model_url_config.yml', 'w') as file:
            documents = yaml.dump(outputs, file)

    def add_file_to_oss(self, src, dist, outputs):
        for file_name in os.listdir(src):
            new_dist = os.path.join(dist, file_name)
            new_src = os.path.join(src, file_name)
            if os.path.isdir(new_src):
                self.add_file_to_oss(new_src, new_dist, outputs)
            else:
                result = self.bucket.put_object_from_file(new_dist, new_src, progress_callback=self.percentage)
                assert result.status == 200
                crc64 = result.headers['x-oss-hash-crc64ecma']
                file_url = os.path.join(HOST_URL, new_dist).__str__()
                local_path = new_src
                outputs.append({
                    "local_path": local_path,
                    "file_url": file_url,
                    "crc": crc64,
                })

    @staticmethod
    def percentage(consumed_bytes, total_bytes):
        if total_bytes:
            rate = int(100 * (float(consumed_bytes) / float(total_bytes)))
            print('\r{0}% '.format(rate), end='')
            sys.stdout.flush()

    def rm_exist_version(self, version_dir):
        for obj in oss2.ObjectIterator(self.bucket, prefix=version_dir):
            self.bucket.delete_object(obj.key)
